#
# Copyright (c) 2010 Nokia Corporation and/or its subsidiary(-ies).
# All rights reserved.
# This component and the accompanying materials are made available
# under the terms of "Eclipse Public License v1.0"
# which accompanies this distribution, and is available
# at the URL "http://www.eclipse.org/legal/epl-v10.html".
#
# Initial Contributors:
# Nokia Corporation - initial contribution.
#
# Contributors:
#
# Description:
# Users configuration manager
#

'''
The filesystem hierarchy for user metadata and corresponding classes. n is a
directory for the nth workspace.

+.blocks (UserConfiguration)
  - blocks.conf (GlobalConfiguration)
  - workspaces (WorkspaceList)
  + n (Workspace)
    - available
    - blocks (BlocksFile)
    - list
    - status
    + info
    + updates

A Blocks workspace is essentially a deb repository. There may be many
workspaces per user on a system.

Integrity

There is no guarantee that workspaces and metadata are consistent if either the
metadata or workspaces are modified simultaneously by multiple clients or
offline by other means.

Orphaned workspace directories are ignored.

The UserConfiguration.removeCruft() method is provided to clean up metadata.

Usage

Creation:
u = UserConfiguration()
u.createWorkspace("foo", "/my/workspaces/foo", ["http://sources.com/sources/foo"])

Access:
v = UserConfiguration()
v.getWorkspaceByName("foo").path
v.getWorkspaceByName("foo").sources
v.getWorkspaceById(1).name


'''

# TODO: This module needs some refactoring

import os
import sys
import shutil
import logging
from xml.dom import minidom
from collections import namedtuple
import platform

# TODO: Get rid of XMLConfigFile and use XMLConfig
from XMLConfigFile import *
from XMLConfig import XMLConfig, Property, ConfigError
if not ".." in sys.path:
    sys.path.append("..")
import utils

class BadWorkspace(Exception):
    ''' The workspace is not listed in metadata '''

class MissingWorkspace(Exception):
    ''' The workspace directory cannot be found '''

WorkspaceConfigError = ConfigError

class WorkspaceList(XMLConfigFile):
    '''
    The I{workspace} configuration file found in the user blocks metadata
    directory.

    int(workspaceID) - str(workspacename) pairs saved as an XML file
    '''
    schema = os.path.join(os.path.dirname(__file__), "workspacelist.wsd")
    WorkspaceInfo = namedtuple("workspace", "id, name, path")

    def __init__(self, configFilePath):
        self._workspaces = {}
        self.default_workspace = 0
        XMLConfigFile.__init__(self, configFilePath)

##        self.name = doc.getElementsByTagName("name")[0].firstChild.nodeValue.strip()
##        self.path = doc.getElementsByTagName("path")[0].firstChild.nodeValue.strip()
    def fromXMLString(self, xmlString):
        doc = minidom.parseString(xmlString)

        default_tag = doc.getElementsByTagName("default")
        if default_tag:
            self.default_workspace = int(default_tag[0].firstChild.nodeValue.strip())
        logging.debug("Default workspace: %s", self.default_workspace)

        for w in doc.getElementsByTagName("workspace"):
            id = int(w.getElementsByTagName("id")[0].firstChild.nodeValue.strip())
            name = w.getElementsByTagName("name")[0].firstChild.nodeValue.strip()
            path = w.getElementsByTagName("path")[0].firstChild.nodeValue.strip()
            wsInfo = self.WorkspaceInfo(id, name, path)
            self._workspaces[id] = wsInfo
            #self._workspaces[name] = wsInfo

    def toXMLString(self):
        doc = minidom.Document()
        workspaces = doc.createElement("workspaces")

        default = doc.createElement("default")
        default.appendChild(doc.createTextNode(str(self.default_workspace)))
        workspaces.appendChild(default)

        doc.appendChild(workspaces)
        for id, ws in self._workspaces.iteritems():
            n = doc.createElement("name")
            n.appendChild(doc.createTextNode(ws.name))
            i = doc.createElement("id")
            i.appendChild(doc.createTextNode(str(id)))
            p = doc.createElement("path")
            p.appendChild(doc.createTextNode(ws.path))
            w = doc.createElement("workspace")
            w.appendChild(n)
            w.appendChild(i)
            w.appendChild(p)
            workspaces.appendChild(w)
        return doc.toprettyxml()

    def setDefaultWorkspace(self, id):
        try:
            id = int(id)
        except ValueError:
            pass
        if id in self.getWorkspaces() + [0]: # Use zero id to unset
            self.default_workspace = id
            self.write()
        else:
            raise BadWorkspace("No workspace with id '%s'" % id)

    def getWorkspaceIdByName(self, name):
        #return self._workspaces.get(name)
        ids = [v.id for v in self._workspaces.values() if v.name == name]
        return ids[0] if ids else None

    def getWorkspaceNameById(self, id):
        ws = self._workspaces.get(id)
        return ws.name if ws else None

    def getWorkspacePath(self, id):
        ws = self._workspaces.get(id)
        return ws.path if ws else None

    def getNextId(self):
        try:
            keys = self._workspaces.keys()
            keys.sort()
            id = keys[-1] + 1
        except IndexError:
            id = 1
        return id

    def getWorkspaces(self):
        return self._workspaces.keys()

    def addWorkspaceToList(self, name, path, id=None):
        if id == None: # pragma: no cover
            id = self.getNextId()
        if self.getWorkspaceIdByName(name):
            raise BadWorkspace("Workspace name is not unique")
        if len(name) < 1:
            raise BadWorkspace("Name must be a non-empty string")
        if id < 1: # pragma: no cover
            raise BadWorkspace("Id must be a positive integer")

        if not self.isWorkspacePathUnique(path):
            raise BadWorkspace("Workspace path is not unique. Any of the workspace paths must not be under existing workspace path.")

        self._workspaces[id] = self.WorkspaceInfo(id, name, path)
        self.write()

    def isWorkspacePathUnique(self, path):
        for wsid in self.getWorkspaces():
            if not utils.pathsUnique(path, self.getWorkspacePath(wsid)):
                return False
        return True

    def removeWorkspace(self, id):
        idnum = int(id)
        if idnum in self._workspaces:
            del self._workspaces[idnum]
            # If default workspace -> unset default workspace
            if self.default_workspace == idnum:
                logging.debug("Removing workspace which is default. Unsetting default workspace.")
                self.default_workspace = 0
            self.write()

class WorkspaceConfig(XMLConfig):

    def __init__(self, path):
        XMLConfig.__init__(self, path)
        self.addProperty(Property("cache-bundles", Property.BOOLEAN, False))
        self.load()

class Workspace(WorkspaceConfig):
    '''
    Workspace metadata directory; one is created in the user metadata directory
    for each workspace.
    '''
    def __init__(self, metadataDir, name, path):
        '''
        Create a new workspace by specifying metadata directory, workspace name and workspace path.

        OR

        Read an existing Workspace by specifying metadata path only.

        @param metadataDir:       Location of workspace metadata directory
        @type metadataDir:        String
        @param name:     Name of the workspace
        @type name:      String
        @param path:     Location of the workspace
        @type path:      String
        '''
        self.metadataDir = metadataDir
        if not os.path.isdir(self.metadataDir):
            path = os.path.abspath(path)
            os.mkdir(self.metadataDir)
            for d in ("info", "updates", "triggers"):
                os.mkdir(os.path.join(self.metadataDir, d))
            for f in ("available", "list", "status"):
                h = open(os.path.join(self.metadataDir, f), "w")
                h.close()
            # Apt dirs
            aptBaseDir = os.path.join(self.metadataDir, "apt")
            os.makedirs(os.path.join(aptBaseDir, "lists", "partial"))
            os.makedirs(os.path.join(aptBaseDir, "cache", "archives", "partial"))
            # TODO: Copy apt config

        self.name = name
        self.path = os.path.normcase(path)

        WorkspaceConfig.__init__(self, os.path.join(self.metadataDir, "blocks.conf"))

class UserConfiguration(WorkspaceList, WorkspaceConfig):
    '''
    The user blocks metadata.
    @var path: The name of the directory containing the metadata files.
    '''
    def __init__(self):
        self.path = utils.getMetaPath()
        utils.createFile(os.path.join(self.path, "trusted.gpg"))

        WorkspaceList.__init__(self, os.path.join(self.path, "workspaces"))
        WorkspaceConfig.__init__(self, os.path.join(self.path, "blocks.conf"))

        self.logger = logging.Logger(self.__class__.__name__)

    def workspaceExists(self, id):
        return True if self.getWorkspaceNameById(id) else False

    def checkWorkspaceExists(self, id):
        if not self.getWorkspaceNameById(id):
            raise BadWorkspace("Workspace with id %s does not exist." % id)

    def getWorkspaceMetadataPath(self, id):
        return os.path.join(self.path, str(id))

    def createWorkspace(self, name, path):
        '''
        @param name:    workspace name
        @type name:     String
        @param path:    workspace location
        @type path:     String
        '''
        # check name is unique
        # get next ID
        # create directory and empty contents
        # write to workspaces file
        id = self._getNextDirectoryId(self.path)
        if self.getNextId() > id:
            id = self.getNextId()
        mdpath = self.getWorkspaceMetadataPath(id)
        while os.path.isfile(mdpath) or os.path.isdir(mdpath): # pragma: no cover
            id += 1
            mdpath = self.getWorkspaceMetadataPath(id)
        self.addWorkspaceToList(name, path, id)
        Workspace(mdpath, name, path)

    @staticmethod
    def _getNextDirectoryId(path):
        '''
        Get the next ID in sequence.

        I{path} is taken to contain files and/or subdirectories named with a
        running sequence of integers

        @return: version
        @rtype: Integer
        '''
        try:
            contents = [int(d) for d in os.listdir(path)]
            contents.sort()
            last = int(contents[-1]) + 1
            return last
        except Exception:
            return 1 # pragma: no cover

    def getWorkspaceById(self, id):
        wsname = self.getWorkspaceNameById(id)
        if not wsname:
            raise BadWorkspace("Workspace with id '%s' does not exist" % id)
        return Workspace(self.getWorkspaceMetadataPath(id), wsname, self.getWorkspacePath(id))

    def getWorkspaceByName(self, name): # pragma: no cover
        id = self.getWorkspaceIdByName(name)
        if not id:
            raise BadWorkspace("Workspace with name '%s' does not exist" % name)
        return Workspace(self.getWorkspaceMetadataPath(id), self.getWorkspaceNameById(id), self.getWorkspacePath(id))

    def getOrphanedListEntries(self):
        ''' Workspaces in list that have no metadatadir '''
        return [wid for wid in self.getWorkspaces() if not os.path.isdir(self.getWorkspaceMetadataPath(wid))]

    def getOrphanedMetadataDirs(self):
        ''' Metadata directories that have no corresponding list entry '''
        return [os.path.join(self.path, d)
                for d in os.listdir(self.path)
                if os.path.isdir(os.path.join(self.path, d)) and d.isdigit() and
                int(d) not in self.getWorkspaces()]

    def getOrphanedWorkspaces(self):
        ''' Workspaces that no longer exist. Ignore errors caused missing metadata. '''
        orphaned = []
        for wid in self.getWorkspaces():
            try:
                if not os.path.isdir(self.getWorkspaceById(wid).path):
                    orphaned.append(wid)
            except UserConfiguration.MissingWorkspace:
                pass
        return orphaned

    def removeWorkspace(self, id):
        '''
        @param id: WS ID
        '''
        path = self.getWorkspaceMetadataPath(id)
        WorkspaceList.removeWorkspace(self, id)
        if os.path.isdir(path):
            shutil.rmtree(path)

    def removeCruft(self, checkForBlocks=True): # pragma: no cover
        ''' Delete workspaces from list that have no corresponding metadatadir '''
        for gone in self.getOrphanedListEntries():
            self.logger.info("Removing workspace %s" % gone)
            self.removeWorkspace(gone)

        for gone in self.getOrphanedMetadataDirs():
            if checkForBlocks:
                if not os.path.isfile(os.path.join(gone, "blocks")):
                    self.logger.warning("No blocks file in metadatadir, not removing %s"%gone)
                    return
            self.logger.info("Removing metadata dir %s" % gone)
            shutil.rmtree(gone)

        for gone in self.getOrphanedWorkspaces():
            self.logger.info("Removing workspace %s" % gone)
            self.removeWorkspace(gone)
