#
# Copyright (c) 2010 Nokia Corporation and/or its subsidiary(-ies).
# All rights reserved.
# This component and the accompanying materials are made available
# under the terms of "Eclipse Public License v1.0"
# which accompanies this distribution, and is available
# at the URL "http://www.eclipse.org/legal/epl-v10.html".
#
# Initial Contributors:
# Nokia Corporation - initial contribution.
#
# Contributors:
#
# Description:
# Simple XML-configuration file with multiple properties
#

import os
import sys
if ".." not in sys.path:
    sys.path.append("..")

import elementtree as ET
import utils

class Property(object):
    '''Property

    If value is None, default value will be returned as value
    '''
    STRING, BOOLEAN = range(2)

    def __init__(self, name, type, value):
        assert type == Property.STRING or type == Property.BOOLEAN
        self.name = name
        self.type = type
        self.value = value

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        if self.type == Property.BOOLEAN:
            origValue = value
            value = utils.toBoolean(value)
            if value is None:
                raise ValueError("Trying to set invalid value '%s' for boolean property '%s'" % (origValue, self.name))
        self._value = value

    def setValue(self, value):
        ''' Convenience method to set value and get (possibly converted) value back '''
        self.value = value
        return self.value

class ConfigError(Exception):
    ''' Configuration Error '''

# TODO: Should properties be case-insensitive?
class XMLConfig(object):
    ''' Simple XML-configuration file with multiple properties '''
    def __init__(self, path):
        '''Opens xml-file

        After initialization add properties into config and load
        '''
        self.configpath = path
        if not os.path.isfile(path):
            self._create()

        self.tree = ET.MyElementTree(file=path)
        self.properties = {}

    def addProperty(self, prop):
        '''Adds property to config so it can be set/get

        Give Property instance as prop.
        '''
        if not isinstance(prop, Property):
            raise ValueError("prop argument must be Property instance")
        self.properties[prop.name] = prop

    def removeProperty(self, name):
        self._checkPropertyName(name)
        del self.properties[name]

    def load(self):
        for prop in self.properties.itervalues():
            value = self.tree.findtext(prop.name)
            self.properties[prop.name].loaded = False
            if value is not None:
                self.properties[prop.name].value = value
                self.properties[prop.name].loaded = True

    def getPropertyValue(self, name, getDefault=True):
        self._checkPropertyName(name)
        if not getDefault and not self.properties[name].loaded:
            return None
        return self.properties[name].value

    def setPropertyValue(self, name, value):
        ''' Set property value '''
        self._checkPropertyName(name)
        try:
            value = str(self.properties[name].setValue(value))
        except ValueError, ex:
            raise ConfigError(str(ex))
        tag = self.tree.find(name)
        if tag is None:
            ET.SubElement(self.tree.getroot(), name).text = value
        else:
            tag.text = value
        self._write()

    def getPropertyValues(self):
        ''' Get all the properties available '''
        values = {}
        for prop in self.properties.itervalues():
            values[prop.name] = prop.value
        return values

    def _checkPropertyName(self, name):
        if name not in self.properties:
            raise ConfigError("Invalid property name '%s'" % name)

    def _create(self):
        self.tree = ET.MyElementTree(ET.Element("properties"))
        self._write()

    def _write(self):
        self.tree.write(self.configpath)

def test(): # pragma: no cover
    config = XMLConfig(r"c:\temp\test.xml")
    config.addProperty(Property("settingx", Property.BOOLEAN, False))
    config.addProperty(Property("settingy", Property.BOOLEAN, False))
    config.setPropertyValue("settingx", "yes")
    print "setting x:", config.getPropertyValue("settingx")
    print "setting y:", config.getPropertyValue("settingy")
    config.setPropertyValue("settingx", "no")
    print "setting x:", config.getPropertyValue("settingx")
    config.setPropertyValue("settingx", "a")

if __name__ == "__main__": # pragma: no cover
    test()