#
# Copyright (c) 2010 Nokia Corporation and/or its subsidiary(-ies).
# All rights reserved.
# This component and the accompanying materials are made available
# under the terms of "Eclipse Public License v1.0"
# which accompanies this distribution, and is available
# at the URL "http://www.eclipse.org/legal/epl-v10.html".
#
# Initial Contributors:
# Nokia Corporation - initial contribution.
#
# Contributors:
#
# Description:
# Command line command dispatcher base class
#

import inspect
import logging
from optparse import OptionParser
import pprint
from collections import namedtuple

import utils

ParameterInfo = namedtuple("ParameterInfo", "args, varargs, defaults, minArgCount, maxArgCount, cmd")

class CommandOptionParser(OptionParser):
    ''' Option parser for commands '''

    def __init__(self, *args, **kwargs):
        OptionParser.__init__(self, usage="", add_help_option=False, *args, **kwargs)

    def error(self, msg):
        raise CmdLineCommand.ArgumentError(msg)

class CmdLineCommand(object):
    ''' Run commands '''

    COMMAND_PREFIX = "cmd_"
    ADVANCED_COMMAND_PREFIX = COMMAND_PREFIX + "advanced_"

    INFINITE = 99999

    class ArgumentError(Exception):
        """ Wrong argument count for command """

    class CommandNotAvailableError(Exception):
        """ Command not available """

    def __init__(self, options):
        utils.initSearchPath()
        self.options = options
        self.cmd_options = {}

    @classmethod
    def getUsage(cls, advanced=False):
        usage = "%prog [OPTIONS] COMMAND [COMMAND-OPTIONS] [ARGS]\n\n"
        commands = cls.getCommands()
        cmdMaxLen = max(len(cmd) for cmd in commands) - len(cls.COMMAND_PREFIX)

        advCommands = cls.getCommands(True)
        if advanced:
            if advCommands:
                advCmdMaxLen = max(len(cmd) for cmd in advCommands) - len(cls.ADVANCED_COMMAND_PREFIX)
                cmdMaxLen = max(cmdMaxLen, advCmdMaxLen)

        if advCommands:
            usage += "Commonly used commands:"
        else:
            usage += "Commands:"
        usage += cls.formatCmdHelp(commands, cmdMaxLen)

        if advanced and advCommands:
            usage += "\n\nAdvanced commands:"
            usage += cls.formatCmdHelp(advCommands, cmdMaxLen)

        return usage

    def cmd_help(self, command):
        ''' Show help for specific command '''
        print self.getCommandUsage(command)

    def __call__(self, args):
        info = self._getCmdParameterInfo(args[0])
        if info:
            arguments = args[1:]
            cmd = info.cmd
            if hasattr(cmd, "_optionparser"):
                (self.cmd_options, arguments) = cmd._optionparser.parse_args(arguments)
                logging.debug("Command options:\n%s", pprint.pformat(self.cmd_options.__dict__))

            argumentCount = len(arguments)
            if argumentCount < info.minArgCount:
                raise self.ArgumentError("Not enough arguments for command '%s'" % args[0])
            elif argumentCount > info.maxArgCount:
                raise self.ArgumentError("Too many arguments for command '%s'" % args[0])
            else:
                cmd(self, *arguments)
        else:
            raise self.CommandNotAvailableError("Command '%s' not available" % args[0])

    @classmethod
    def _getCmdParameterInfo(cls, command):
        cmd = getattr(cls, cls._cmdNameToInternalName(command), None)
        if cmd and callable(cmd):
            # Check function parameters
            argspec = inspect.getargspec(cmd)
            assert argspec.keywords == None, "no keyword arguments allowed"
            # self not "real" parameter. substract 1
            posArgCount = len(argspec.args) - 1
            defArgCount = len(argspec.defaults) if argspec.defaults else 0
            minArgCount = posArgCount - defArgCount
            # Infinite for variable parameter functions
            maxArgCount = posArgCount if not argspec.varargs else cls.INFINITE
            return ParameterInfo(argspec.args[1:], argspec.varargs, argspec.defaults, minArgCount, maxArgCount, cmd)
        else:
            return False

    #TODO: Add automatic alias generation?
    @classmethod
    def _cmdNameToInternalName(cls, command):
        command = command.replace("-", "_")
        cmdName = cls.COMMAND_PREFIX + command
        advCmdName = cls.ADVANCED_COMMAND_PREFIX + command
        if hasattr(cls, cmdName):
            command = cmdName
        elif hasattr(cls, advCmdName):
            command = advCmdName
        return command

    @classmethod
    def getCommandUsage(cls, command):
        info = cls._getCmdParameterInfo(command)
        if info:
            doc = inspect.getdoc(info.cmd)
            # If doc string is multi-line string add new-line after it
            if "\n" in doc:
                doc += "\n"
            # Get full doc string from method without extra line break after summary
            doc = doc.replace("\n", "", 1).strip()
            usage = "Purpose: %s\nUsage: %s" % (doc, command)

            cmd = getattr(cls, cls._cmdNameToInternalName(command))
            optionParser = getattr(cmd, "_optionparser", None)
            if optionParser:
                usage += " [options]"

            for arg in info.args:
                usage += " <%s>" % arg
            if info.varargs:
                usage += " [%s%d ...]" % (info.varargs, 2 if info.args else 1)
            if optionParser:
                usage += "\n\n" + optionParser.format_help().strip()
            return usage
        else:
            return "Command '%s' not available" % command

    @classmethod
    def getCommands(cls, advanced=False):
        cmd_list = [n for n in dir(cls) if n.startswith(cls.COMMAND_PREFIX)]
        if advanced:
            return [n for n in cmd_list if n.startswith(cls.ADVANCED_COMMAND_PREFIX)]
        elif advanced == False:
            return [n for n in cmd_list if not n.startswith(cls.ADVANCED_COMMAND_PREFIX)]
        else:
            return cmd_list

    @classmethod
    def getCommandNames(cls):
        for cmd in cls.getCommands(None):
            cmd = cmd.replace("advanced_", "")
            cmd = cmd[4:].replace("_", "-")

    @classmethod
    def formatCmdHelp(cls, commands, cmdMaxLen):
        usage = ""
        for cmd in commands:
            usage += "\n"
            cmdHelp = inspect.getdoc(getattr(cls, cmd))
            # Get the summary line from doc string
            cmdHelp = cmdHelp.splitlines()[0].strip() if cmdHelp else ""
            cmd = cmd.replace("advanced_", "")
            usage += "  %-*s  " % (cmdMaxLen, cmd[4:].replace("_", "-"))
            usage += cmdHelp
        return usage