#
# Copyright (c) 2010 Nokia Corporation and/or its subsidiary(-ies).
# All rights reserved.
# This component and the accompanying materials are made available
# under the terms of "Eclipse Public License v1.0"
# which accompanies this distribution, and is available
# at the URL "http://www.eclipse.org/legal/epl-v10.html".
#
# Initial Contributors:
# Nokia Corporation - initial contribution.
#
# Contributors:
#
# Description:
# comps.xml manager
#

from collections import defaultdict, namedtuple
import logging
import re

import elementtree as ET

Groups = namedtuple("Groups", "Installed, Available")

#TODO: XML validation?

class Comps(object):
    ''' Manages comps.xml files '''

    _GROUP_TAG = "group"

    def __init__(self):
        self._tagProcessors = [self._processGroupTag]
        self._groups = {}

    def create(self, filename_or_obj):
        ''' Create comps file with filename or write to file object '''
        tree = ET.MyElementTree(ET.Element("comps"))
        for group in (g for g in self._groups.itervalues() if len(g.allPackages) > 0):
            groupElem = ET.SubElement(tree.getroot(), "group")
            ET.SubElement(groupElem, "id").text = group.id
            ET.SubElement(groupElem, "name").text = group.name
            ET.SubElement(groupElem, "description").text = group.description
            packagesElem = ET.SubElement(groupElem, "packagelist")
            for packages, packageType in (
            (group.mandatoryPackages, "mandatory"),
            (group.defaultPackages, "default"),
            (group.optionalPackages, "optional")):
                if packageType == "mandatory":
                    for package in packages:
                        ET.SubElement(packagesElem, "packagereq").text = package
                else:
                    for package in packages:
                        ET.SubElement(packagesElem, "packagereq", {"type": packageType}).text = package
        tree.write(filename_or_obj)

    @property
    def groups(self):
        ''' Namedtuple with Installed and Available attributes '''
        return Groups([g for g in self._groups.itervalues() if g.installed],
                      [g for g in self._groups.itervalues() if not g.installed])

    def addGroup(self, group):
        self._groups[group.name] = group

    def getGroup(self, name):
        return self._groups.get(name)

    def getGroupById(self, idname):
        try:
            return [g for g in self._groups.itervalues() if g.id == idname][0]
        except IndexError:
            return None

    def _processGroupTag(self, elem):
        if elem.tag == self._GROUP_TAG:
            group = Group(elem=elem)
            self._groups[group.name] = group
            elem.clear()

    def add(self, srcfile):
        ''' Add groups from comps.xml '''
        try:
            for _, elem in ET.iterparse(srcfile):
                for processor in self._tagProcessors:
                    processor(elem)
        except SyntaxError, ex:
            raise SyntaxError("Parsing of comps file '%s' failed: %s" % (srcfile, ex))
        except IOError, ex:
            raise IOError("Problem on opening comps file: %s" % ex)

    def resolveInstalledGroups(self, installedPackages):
        ''' Resolve installation status for all groups '''
        if installedPackages:
            for group in self._groups.itervalues():
                mandatoryPackages = group.packages[Group.MANDATORY_TYPE]
                if mandatoryPackages:
                    group.installed = mandatoryPackages.issubset(installedPackages)
                else:
                    nonMandatoryPackages = group.packages[Group.DEFAULT_TYPE] + group.packages[Group.OPTIONAL_TYPE]
                    group.installed = not nonMandatoryPackages.isdisjoint(installedPackages)

class Group(object):
    ''' Package group '''

    MANDATORY_TYPE = "mandatory"
    DEFAULT_TYPE = "default"
    OPTIONAL_TYPE = "optional"

    def __init__(self, id="", name="", description="", elem=None):
        self.id = id
        self.name = name
        self.description = description
        self.installed = False
        self.packages = defaultdict(set)

        if elem:
            for child in elem:
                if child.tag in ("id", "name", "description"):
                    self.__dict__[child.tag] = child.text
                elif child.tag == "packagelist":
                    for package in child:
                        reqtype = package.get("type")
                        if not reqtype:
                            reqtype = self.MANDATORY_TYPE
                        self.packages[reqtype].add(package.text)

    def addPackages(self, packages, ptype=None):
        '''
        Accepts one package name or list of names
        type is by default mandatory
        '''
        ptype = ptype or self.MANDATORY_TYPE
        assert ptype in (self.MANDATORY_TYPE, self.DEFAULT_TYPE, self.OPTIONAL_TYPE), "Wrong type"
        if not isinstance(packages, list):
            packages = [packages]
        self.packages[ptype].update(packages)

    @property
    def mandatoryPackages(self):
        return list(self.packages[self.MANDATORY_TYPE])

    @property
    def defaultPackages(self):
        return list(self.packages[self.DEFAULT_TYPE])

    @property
    def optionalPackages(self):
        return list(self.packages[self.OPTIONAL_TYPE])

    @property
    def allPackages(self):
        return (self.mandatoryPackages +
                self.defaultPackages +
                self.optionalPackages)

Rule = namedtuple("Rule", "groupid, type, key, regexp")
class CompsTemplate(Comps):
    ''' Parses also rules from comps.xml and processes those with metadata '''

    _RULES_TAG = "rules"

    def __init__(self, filename):
        self.rules = []
        Comps.__init__(self)
        self._tagProcessors.append(self.rulesTagProcessor)
        self.add(filename)

    def applyRule(self, metadata):
        for rule in self.rules:
            value = metadata.get(rule.key)
            if value:
                if re.search(rule.regexp, value):
                    group = self.getGroupById(rule.groupid)
                    if group:
                        packageName = metadata.get("Package")
                        if packageName:
                            group.addPackages(packageName, rule.type)
                        else:
                            raise SyntaxError("Corrupted metadata. No package information available.")
                    else:
                        raise SyntaxError("Comps template: Invalid group id '%s' in rule" % rule.groupid)
            else:
                logging.warning("Comps template: Key '%s' not found from bundle metadata", rule.key)

    def rulesTagProcessor(self, elem):
        if elem.tag == self._RULES_TAG:
            for rule in elem:
                groupId = rule.get("groupid")
                if groupId:
                    packageType = rule.get("type", "mandatory")
                    for match in rule:
                        key = match.get("key", "Package")
                        regexp = match.text
                        self.rules.append(Rule(groupId, packageType, key, regexp))
            elem.clear()