#
# Copyright (c) 2010 Nokia Corporation and/or its subsidiary(-ies).
# All rights reserved.
# This component and the accompanying materials are made available
# under the terms of "Eclipse Public License v1.0"
# which accompanies this distribution, and is available
# at the URL "http://www.eclipse.org/legal/epl-v10.html".
#
# Initial Contributors:
# Nokia Corporation - initial contribution.
#
# Contributors:
#
# Description:
# Component builder
#

'''
ComponentBuilder is the scaffolding for creating PackageModel.Component objects
from a set of files introduced via BuildData. This could/should be a giant
constructor in Component, but the idea was to keep this specific implementation
separate from the generic package model.
'''

import re
import os
import xml.dom.minidom as dom

from SymbianUtils.Evalid import Evalid
from SymbianUtils.Readelf import Readelf
from SymbianUtils import SymbianUtilsError

from PackageModel import *
from Storage import NoPreviousBuild
import Blocks.Packaging.Rules
from Blocks.Packaging.Logging import Logging
from Blocks.Packaging import PackagingError
from Blocks.Packaging.PackageWriter import Dependency

class BuilderException(PackagingError):
    ''' Parent to more specific exceptions. Error in packaging rules, previous metadata or input. '''

class FileError(BuilderException):
    ''' Missing listed file or non-optional derived file '''

class DerivedFileError(BuilderException):
    ''' Missing file derived via rules '''

class BuildHistoryError(BuilderException):
    ''' Previous metadata has errors or does not match current build '''

class ConfigError(BuilderException):
    ''' ConfigError - problem with rules etc. '''

class ComponentBuilder(object):

    def __init__(self, targetRulesPath=None, sourceRulesPath=None, directivesPath=None, keepGoing=True):
        self.targetRules = self.loadPackagingRules(targetRulesPath or Blocks.Packaging.Rules.targetRules())
        self.sourceRules = self.loadPackagingRules(sourceRulesPath or Blocks.Packaging.Rules.sourceRules())
        self.directives = dom.parse(directivesPath or Blocks.Packaging.Rules.packageDirectives())
        self.packages = {}
        self.sources = None
        self.oldComp = None
        self.newComp = None
        self.precLess = []
        self.precMore = []
        self.precSame = []
        self.precDifferent = []
        self.precDone = []
        self.precStarted = []
        self.buildData = None
        self.keepGoing = keepGoing # Ignore all errors

    @staticmethod
    def getLogger():
        ''' Getting the logger each time makes the module picklable '''
        return Logging.getLogger("pfw.componentbuilder")

    def createComponent(self, storage, buildData):
        '''
        The method you want to call to use this class.

        @param buildData: A BuildData containing the files, attributes etc. for
                          the component.
        @type buildData: BuildData
        @return: Component The ready component
        @rtype: Component Component
        @raise BuilderException: Error in packaging rules, previous metadata or input.
        '''
        self.buildData = buildData
        self.packages = {}
        self.sources = None
        self.oldComp = None
        self.newComp = None
        self.precLess = []
        self.precMore = []
        self.precSame = []
        self.precDifferent = []
        self.precDone = []
        self.precStarted = []

        if buildData.getTargetFiles() and buildData.getTargetRoot():
            self.__addDeliverables(buildData.getTargetFiles(), buildData.getTargetRoot())
        else:
            self.getLogger().debug("%s: No target being created. TargetRoot '%s', %s targets.",
                buildData.getComponentName(), buildData.getTargetRoot(), len(buildData.getTargetFiles()))
        if buildData.getSourceFiles() and buildData.getSourceRoot():
            self.__addSources(buildData.getSourceFiles(), buildData.getSourceRoot())
        else:
            self.getLogger().debug("%s: No source package created. SourceRoot '%s', %s sources.",
                buildData.getComponentName(), buildData.getSourceRoot(), len(buildData.getSourceFiles()))

        self.__finalizePackages()
        self.newComp = Component(buildData.getComponentName(), buildData.getComponentVersion(), self.packages.values())
        if self.sources:
            self.newComp.setSourcePackage(self.sources)
        self.newComp.setBuild(storage.getBuildId())
        self.newComp.setRevision(buildData.getComponentRevision())
        self.newComp.getIdentifier().setUseEpoch(buildData.getUseEpoch())

        for name, value in buildData.attributes.items():
            significant = False
            if name in buildData.significantAttributes:
                significant = True
            self.newComp.setAttribute(name, value, significant)

        try:
            self.oldComp = Component.load(storage.getLastMetaDataFile(componentName=self.newComp.getName()))

            if self.newComp.getName() != self.oldComp.getName():
                raise BuildHistoryError("Incompatible component. Cannot compare '%s' with previous build named '%s'" %
                    (self.newComp.getName(), self.oldComp.getName()))

            if buildData.forcePackageAll:
                self.getLogger().debug("%s: packaging forced" % self.newComp.getName())
            else:
                self.getLogger().debug("%s: resolving packaging" % self.newComp.getName())
                # The default is to package all packages. If not forced, this
                # is toggled off and resolvePackaging() sets per package values.
                self.__resolvePackaging()
        except NoPreviousBuild, e:
            self.getLogger().debug(str(e))

        if self.oldComp:
            if self.oldComp.getVersion() == self.newComp.getVersion():
                if self.oldComp.getBuild() != self.newComp.getBuild():
                    self.newComp.setRelease(str(int(self.oldComp.getRelease()) + 1))

        self.__resolveApiVersions()
        self.__setPackageVersions()
        self.__resolveDependencies()
        return self.newComp

    @staticmethod
    def loadPackagingRules(path):
        '''
        Parse XML configuration and create packaging rules.

        @param path: Path of configuration file
        @type path: String
        @todo: validate xml elsewhere and skip checks
        @raise ConfigError: Incomplete rules
        '''
        rules = []
        doc = dom.parse(path)
        for rule in doc.getElementsByTagName("rule"):
            matchPath = ""
            theType = ""
            uids = {}
            package = ""
            variant = None
            extras = []
            optionalExtras = []
            api = None
            try:
                matchPath = rule.getElementsByTagName("path")[0].firstChild.nodeValue.encode("ascii")
                theType = rule.getElementsByTagName("type")[0].firstChild.nodeValue.encode("ascii")
            except Exception:
                raise ConfigError("Missing rule match or type.")
            if not theType:
                raise ConfigError("Missing type in rule.")
            if not matchPath:
                raise ConfigError("Missing match path in rule.")
            if not theType == "ignore":
                # A package is mandatory for non-ignored files
                try:
                    package = rule.getElementsByTagName("package")[0].firstChild.nodeValue.encode("ascii")
                except IndexError:
                    raise ConfigError("Package rule missing 'package' element.")
                # Everything else is optional
                if rule.getElementsByTagName("variant") and rule.getElementsByTagName("variant")[0].firstChild:
                    variant = rule.getElementsByTagName("variant")[0].firstChild.nodeValue.encode("ascii")
                for x in (1, 2, 3):
                    if rule.getElementsByTagName("uid" + str(x)):
                        uids[x] = rule.getElementsByTagName("uid" + str(x))[0].firstChild.nodeValue.encode("ascii")
                extras = [m.firstChild.nodeValue.encode("ascii") for m in rule.getElementsByTagName("mandatory")]
                optionalExtras = [m.firstChild.nodeValue.encode("ascii") for m in rule.getElementsByTagName("optional")]
                if rule.getElementsByTagName("api"):
                    api = rule.getElementsByTagName("api")[0].firstChild.nodeValue.encode("ascii")
            rules.append(Rule(matchPath, theType, uids, package, variant, api, extras, optionalExtras))
        return(rules)

    def __addDeliverables(self, pathList, baseDir):
        '''
        - Use packaging rules to create and store Deliverable object, as well
          as any extras.
        - Add Deliverables to self.packages dictionary where package names are keys.

        @param pathList: Paths relative to baseDir (epocroot)
        @type pathList: List(String)
        @param baseDir: Absolute path to epocroot.
        @type baseDir: String
        @raise BuilderException: Things broke
        @raise FileError: No file found in path
        '''
        for path in pathList:
            absPath = os.path.join(baseDir, path)
            if not os.path.isfile(absPath):
                if self.keepGoing:
                    self.getLogger().warning("Missing listed target file %s" % absPath)
                    continue
                else:
                    raise FileError("Missing listed target file %s" % absPath)
            matched = False
            for rule in self.targetRules:
                try:
                    match, files, extras = rule(path, baseDir, self.keepGoing, absPath, self.buildData.apiMapping)
                    if match:
                        matched = True
                        for pkg, file in files:
                            if isinstance(file, Deliverable):
                                if pkg not in self.packages:
                                    self.packages[pkg] = Package(pkg)
                                self.packages[pkg].addFile(file)
                        for e in extras:
                            self.__addDeliverables([e], baseDir)
                        break
                except SymbianUtilsError, e:
                    if self.keepGoing:
                        self.getLogger().warning(str(e))
                        matched = True # Prevent a log entry saying it did not match any rules
                        break          #
                    else:
                        raise BuilderException(str(e))

            if not matched:
                _e = "%s: deliverable did not match any rules '%s'" % (self.buildData.componentName, path)
                if self.keepGoing:
                    self.getLogger().info(_e)
                else:
                    raise ConfigError(_e)

    def __addSources(self, pathList, baseDir):
        '''
        - Use packaging rules to create and store Deliverable object.
        - Add Sources to source package.

        @param pathList: Paths relative to baseDir (epocroot)
        @type pathList: List(String)
        @param baseDir: Absolute path to epocroot.
        @type baseDir: String
        @raise BuilderException: Things broke
        @raise FileError: No file found in path
        '''
        for path in pathList:
            absPath = os.path.join(baseDir, path)
            if not os.path.isfile(absPath):
                if self.keepGoing:
                    self.getLogger().warning("Missing listed source file %s" % absPath)
                    continue
                else:
                    raise FileError("Missing listed source file %s" % absPath)
            matched = False
            for rule in self.sourceRules:
                try:
                    match, files, extras = rule(path, baseDir, self.keepGoing, absPath)
                    if match:
                        matched = True
                        for pkg, file in files:
                            if isinstance(file, Deliverable):
                                if not self.sources:
                                    self.sources = Package(pkg)
                                self.sources.addFile(file)
                        for e in extras:
                            self.__addSources([e], baseDir)
                        break
                except SymbianUtilsError, e:
                    if self.keepGoing:
                        self.getLogger().warning(str(e))
                        matched = True # Prevent a log entry saying it did not match any rules
                        break          #
                    else:
                        raise BuilderException(str(e))

            if not matched:
                _e = "%s: source did not match any rules '%s'" % (self.buildData.componentName, path)
                if self.keepGoing:
                    self.getLogger().info(_e)
                else:
                    raise ConfigError(_e)

    def __finalizePackages(self):
        '''
        Postprocess and finalize packages, renaming and applying
        dependencies and preconditions using package directives.
        '''
        # First pass get names
        componentName = self.buildData.getComponentName()
        if self.sources:
            self.sources.name = self.resolvePackageName(self.sources.name, componentName, self.directives)
        for k, v in self.packages.items():
            out = self.resolvePackageName(v.name, componentName, self.directives)
            if out:
                # Add possible replaces by using previous names of the component if given
                replaces = [self.resolvePackageName(v.name, n, self.directives) for n in self.buildData.previousNames]
                self.packages[k].replaces = replaces
                self.packages[k].name = out
            else:
                raise Exception("failed to add name to package %s" % k)

        # After this the final names of deps and precs can be resolved
        for k, v in self.packages.items():
            out = self.__getPackageInfo(k)
            if out:
                self.packages[k].arch, \
                self.packages[k].depends, \
                self.packages[k].preconditions = out
                self.getLogger().debug("Applied directives. Name: %s Arch: %s Depends: %s Preconditions: %s"%(
                            self.packages[k].getName(),
                            self.packages[k].arch,
                            str([str(d) for d in self.packages[k].depends]),
                            str(self.packages[k].preconditions))
                        )
            else:
                raise Exception("failed to add data to package %s" % k)

    @staticmethod
    def resolvePackageName(pkgName, componentName, directives):
        '''
        Use packageDirectives to work out full name.

        For example::
            >>> resolvePackageName("exec.arm", "foo", Rules.packageDirectives())
            'foo.exec-arm'

        @param pkgName: The package name as determined by packaging rules
        @return: The full name of the package from component and suffix
        @rtype: String or None
        '''
        for d in directives.getElementsByTagName("rule"):
            rex = re.compile(d.getElementsByTagName("package")[0].firstChild.nodeValue.encode("ascii"))
            mo = rex.match(pkgName)
            if not mo:
                continue
            else:
                if d.getElementsByTagName("suffix")[0].firstChild:
                    fullName = componentName + d.getElementsByTagName("suffix")[0].firstChild.nodeValue.encode("ascii")
                else:
                    fullName = componentName
                for i in range(1, len(mo.groups())+1):
                    fullName = fullName.replace("$%s" % i, mo.group(i))
                return(fullName)
        return None

    def __getPackageInfo(self, pkgName):
        '''
        Use packageDirectives to work out architecture and relations to other
        packages in the component.

        Running this only makes sense once all packages have been added and
        names updated using getPackageName()

        @param pkgName: The package name as determined by packaging rules
        @return: (arch, depends, preconditions).
            - The architecture of the package
            - Packages this depends on
            - list of Package names
        @rtype: String, List(Dependency), List(String)
        '''
        for d in self.directives.getElementsByTagName("rule"):
            rex = re.compile(d.getElementsByTagName("package")[0].firstChild.nodeValue.encode("ascii"))
            mo = rex.match(pkgName)
            if not mo:
                continue
            else:
                arch = d.getElementsByTagName("arch")[0].firstChild.nodeValue.encode("ascii")
                deps = []
                preconditions = []
                try:
                    deps = [Dependency(m.firstChild.nodeValue.encode("ascii"), m.getAttribute("type").encode("ascii"))
                            for m in d.getElementsByTagName("depends")]
                except IndexError:
                    pass
                try:
                    preconditions = [m.firstChild.nodeValue.encode("ascii") for m in d.getElementsByTagName("precondition")]
                except IndexError:
                    pass
                for i in range(1, len(mo.groups())+1):
                    arch = arch.replace("$%s" % i, mo.group(i))
                    for dep in deps:
                        dep.package = dep.package.replace("$%s" % i, mo.group(i))
                    preconditions = [prec.replace("$%s" % i, mo.group(i)) for prec in preconditions]
                deps = [d for d in deps if d.package in self.packages]
                for dep in deps:
                    dep.package = self.packages[dep.package].name
                preconditions = [self.packages[p].name for p in preconditions if p in self.packages]
                return(arch, deps, preconditions)
        return None

    def __resolvePackaging(self):
        '''
        Toggle package flags in newComp packages based on oldComp

         - No significant changes (checksums are the same, significant metadata
           same): No repackaging
         - Significant component metadata changed: All packages are written
         - If preconditions, write package only if there are significant changes
           in packages listed in preconditions
        '''
        # source package always compared:
        if self.newComp.getSourcePackage():
            if self.oldComp.getSourcePackage():
                if self.newComp.getSourcePackage() == self.oldComp.getSourcePackage():
                    self.newComp.getSourcePackage().setPackageFlag(False)
                    self.getLogger().debug("%s: source package unchanged", self.newComp.getName())
                else:
                    self.getLogger().debug("%s: source package has changed", self.newComp.getName())
            else:
                self.getLogger().debug("%s: source package has been added", self.newComp.getName())
        if self.newComp == self.oldComp:
            self.getLogger().debug("identical component: %s", self.newComp.getName())
            self.newComp.setPackageAllFlag(False)
            for p in self.newComp.getPackages():
                p.setPackageFlag(False)
            return
        # force writing of all packages if vital metadata has changed
        elif self.newComp.getSignificantAttributes() != self.oldComp.getSignificantAttributes():
            self.getLogger().debug("vital metadata changed, force repackage: %s", self.newComp.getName())
            self.newComp.setPackageAllFlag(True)
            return
        # resort to package by package
        else:
            self.newComp.setPackageAllFlag(False)
            self.precLess, self.precMore, self.precSame, self.precDifferent = self.newComp.diff(self.oldComp)
            # different packages repackaged only if a precondition has changed
            for d in self.precDifferent:
                self.getLogger().debug("check changed package for preconditions: %s"%d)
                self.__resolvePreconditions(d)
            for p in self.precSame:
                self.newComp.getPackage(p).setPackageFlag(False)

    def __resolvePreconditions(self, package):
        '''
        A precondition is another package. A package may have preconditions,
        indicating that it may not have to be rewritten even though it has
        changed - only if any precondition has changed.

        Assign to self.precSame those packages that have preconditions and none
        of the packages listed as preconditions has changed. Preconditions can
        also have preconditions, and the relationship can be circular.

        @param package: The name of a package
        '''
        if package in self.precStarted:
            self.getLogger().debug("Preconditions already started: %s", package)
            return
        else:
            self.precStarted.append(package)
        if self.newComp.getPackage(package).preconditions:
            for p in self.newComp.getPackage(package).preconditions:
                self.getLogger().debug("Prec %s: %s", package, p)
                self.__resolvePreconditions(p)
            preconditionsWithChanges = [pre for pre in self.newComp.getPackage(package).preconditions if pre not in self.precSame]
            if not preconditionsWithChanges:
                self.getLogger().debug("%s: preconditions not changed, tagging as unchanged", package)
                self.precSame.append(package)
                self.precDifferent = [d for d in self.precDifferent if d != package]

    def __resolveApiVersions(self):
        ''' Set or update file and package API versions. '''
        for p in self.newComp.getPackages():
            filesWithApi = [f for f in p.getFiles() if (isinstance(f, Executable) and f.api)]
            if filesWithApi:
                if self.oldComp and self.oldComp.getPackage(p.getName()):
                    oldP = self.oldComp.getPackage(p.getName())
                    if oldP.getApiVersion():
                        oldMajor, oldMinor = oldP.getApiVersion().split(".")
                    else:
                        self.getLogger().warning("Previous version of %s has no API version", p.getName())
                        oldMajor = "1"
                        oldMinor = "0"
                    majorChange = False
                    minorChange = False
                    for f in filesWithApi:
                        oldF = oldP.getFile(f.getPath())
                        if oldF and hasattr(oldF, "api"):
                            try:
                                fOldMajor, fOldMinor = oldF.api.version.split(".")
                                assert (int(fOldMajor) > 0)
                                assert (int(fOldMinor) >= 0)
                            except AssertionError:
                                self.getLogger().warning("Previous version of %s has bad API version", f.getPath())
                                fOldMajor = "1"
                                fOldMinor = "0"
                            if f.api.isIdentical(oldF.api):
                                continue
                            elif f.api.provides(oldF.api):
                                f.api.version = (fOldMajor + "." + str(int(fOldMinor) + 1))
                                minorChange = True
                                continue
                            else:
                                f.api.version = (str(int(fOldMinor) + 1) + ".0")
                                majorChange = True
                        else:
                            self.getLogger().warning("Previous version of %s has no API version", f.getPath())
                            # the first API version will be 1.0 by default
                    if majorChange:
                        p.setApiVersion(str(int(oldMajor) + 1) + ".0")
                    elif minorChange:
                        p.setApiVersion(oldMajor + "." + str(int(oldMinor) + 1))
                    else:
                        p.setApiVersion(oldMajor + "." + oldMinor)
                else:
                    if not p.getApiVersion():
                        p.setApiVersion("1.0")

    def __setPackageVersions(self):
        if self.newComp.getSourcePackage():
            p = self.newComp.getSourcePackage()
            p.setIdentifier(self.newComp.getIdentifier())
        for p in self.newComp.getPackages():
            # Inherit version and build id from parent component
            p.setIdentifier(self.newComp.getIdentifier())
            if p.getPackageFlag():
                p.setLastPackagedVersion(p)
            else:
                p.setLastPackagedVersion(self.oldComp.getPackage(p.getName()))

    def __resolveDependencies(self):
        '''
        Each dependency is given a version and evaluator.

         - If the required package is being packaged in this release, the
           dependency is on this version.
         - If a dependency is not packaged this time, get the version from
           previous package metadata.

        The evaluation operator is ">=" if a version is defined, but for now we
        always get the version - we even raise an exception if the version has
        not been set.
        '''
        eval = ">="
        for p in self.newComp.getPackages():
            # dep is a Dependency, currently only holds the name of a Package
            # that package p depends on. See __getPackageInfo().
            for dep in p.depends:
                depPkg = self.newComp.getPackage(dep.package)
                if depPkg.getPackageFlag():
                    # It's being packaged in this release so we require the latest version
                    dep.setIdentifier(depPkg.getIdentifier())
                else:
                    # Not packaged now, use old metadata to get version info
                    if not self.oldComp:
                        raise BuildHistoryError("%s: error while resolving dependencies: metadata for previous builds not available" % self.newComp.getName())
                    depPkg = self.oldComp.getPackage(dep.package)
                    dep.setIdentifier(depPkg.getLastPackagedVersion())
                if (not dep.getVersion()) or (not dep.getBuild()):
                    raise BuildHistoryError("%s: unable to resolve version of dependency: %s -> %s" %
                        (self.newComp.getName(), p.getName(), dep.package))
                dep.setEval(eval)

class Rule(object):
    """
    A set of criteria used to create Deliverable or subclass objects.

    When a Rule object is called it acts as a Deliverable factory:

    r = Rule(criteria)
    deliverable = r(path, baseDirectory)
    """

    def __init__(self, matchPath, type, matchUids=None, package="", variant=None, api=None, extras=None, optionalExtras=None):
        '''
        @param matchPath: Regex that matches the file name and path
        @param matchUids: UIDs
        @param type:
        @param package:
        @param variant:
        @param api: Path to related .dso file
        @param extras:
        @param optionalExtras:
        '''
        self.matchPath = matchPath
        self.pathRe = re.compile(matchPath)
        self.matchUids = matchUids or {}
        self.type = type
        self.package = package
        self.variant = variant
        self.api = api
        self.extras = extras or []
        self.optionalExtras = optionalExtras or []

    @staticmethod
    def getLogger():
        ''' Getting the logger each time makes the module picklable '''
        return Logging.getLogger("pfw.componentbuilder.rule")

    def __call__(self, path, baseDir="", keepGoing=True, absPath=None, apiMapping=None):
        """
        @param path: String
        @param baseDir: String
        @param keepGoing: Ignore DerivedFileError and ConfigError. FileError is never ignored.
        @param absPath: baseDir and path joined (optional)
        @return: isAMatch, files, extras
        @rtype: Boolean, List((packageName, Deliverable)), List(String)
        @raise DerivedFileError: A file derived from rules is missing, excluding optional extras.
        @raise ConfigError: The rule specifies an unrecognized Deliverable type
        @raise SymbianUtils.SymbianUtilsError: A SymbianUtils tool failed while trying to process a file
        """
        if not path:
            raise ValueError("Cannot create a Deliverable from an empty path")
        if absPath is None:
            absPath = os.path.join(baseDir, path)
        apiMapping = apiMapping or {}
        isAMatch = False
        files = [] # contains (packageName, Deliverable) tuples
        extras = []
        m = self.pathRe.match(path) # Path matches
        u = True                    # UIDs match
        if m:
            for k, v in self.matchUids.items():
                u = u and Evalid.getUid(k, absPath) == eval("0x" + str(v))
        if m and u:
            isAMatch = True
            if not self.type == "ignore":
                # replace variables in rules to obtain derived values
                derivedPackage = self.package
                derivedVariant = self.variant
                apiPath = apiMapping.get(path) or apiMapping.get(os.path.basename(path))
                if apiPath:
                    derivedApi = apiPath
                else:
                    derivedApi = self.api           # path to relevant dso file
                for i in range(1, len(m.groups())+1):
                    if m.group(i) == None:
                        continue
                    derivedPackage = derivedPackage.replace("$%s"%i, m.group(i))
                    if derivedVariant:
                        derivedVariant = derivedVariant.replace("$%s"%i, m.group(i))
                    if derivedApi:
                        derivedApi = derivedApi.replace("$%s"%i, m.group(i))
                checksum = Evalid.generateSignature(absPath, self.type)
                theApi = None
                if derivedApi:
                    try:
                        theApi = API(interface=Readelf.getInterface(os.path.join(baseDir, derivedApi)))
                    except Exception, e:
                        _e = "Failed to get API for %s. Rule: '%s', error: %s" % (path, self.matchPath, str(e))
                        print _e
                        if keepGoing:
                            self.getLogger().warning(_e)
                        elif _e.endswith("No such file"):
                            # don't raise exception if missing file
                            self.getLogger().warning(_e)
                        else:
                            # wrong filetype etc. still cause exception
                            # preserve traceback while modifying message
                            _args = list(e.args)
                            _args[0] = _e
                            e.args = tuple(_args)
                            raise

                if self.type in ("exe", "plugin", "dll"):
                    files.append((derivedPackage, Executable(path, self.type, os.path.getsize(absPath), checksum, derivedVariant, theApi)))
                elif self.type in ("staticlib", "dso"):
                    files.append((derivedPackage, Library(path, self.type, os.path.getsize(absPath), checksum, derivedVariant)))
                elif self.type in ("file", "preprocessed_text"):
                    files.append((derivedPackage, Deliverable(path, self.type, os.path.getsize(absPath), checksum)))
                elif self.type in ("intel_pe", "intel"):
                    files.append((derivedPackage, PEBinary(path, self.type, os.path.getsize(absPath), checksum)))
                else:
                    if keepGoing:
                        self.getLogger().error("Unrecognized deliverable type '%s' in rule '%s', adding as Deliverable", self.type, self.matchPath)
                        files.append((derivedPackage, Deliverable(path, self.type, os.path.getsize(absPath), checksum)))
                    else:
                        raise ConfigError("Unrecognized deliverable type '%s' in rule '%s'" % (self.type, self.matchPath))

                for extra in self.extras:
                    for i in range(1, len(m.groups())+1):
                        if m.group(i) == None:
                            continue
                        extra = extra.replace("$%s"%i, m.group(i))
                    if os.path.isfile(os.path.join(baseDir, extra)):
                        extras.append(extra)
                    else:
                        _e = "Missing required file %s (derived from %s) in rule '%s'" % (extra, path, self.matchPath)
                        if keepGoing:
                            self.getLogger().warning(_e)
                        else:
                            raise DerivedFileError(_e)

                for extra in self.optionalExtras:
                    for i in range(len(m.groups())+1):
                        if m.group(i) == None:
                            continue
                        extra = extra.replace("$%s" % i, m.group(i))
                    if os.path.isfile(os.path.join(baseDir, extra)):
                        extras.append(extra)
                    else:
                        self.getLogger().debug("Missing optional file %s (derived from %s)", extra, path)
        return isAMatch, files, extras

    def __str__(self):
        thestr = """
        path: %s
        uid: %s
        package: %s
        """ % (self.matchPath, self.matchUids, self.package)
        return thestr