#
# Copyright (c) 2010 Nokia Corporation and/or its subsidiary(-ies).
# All rights reserved.
# This component and the accompanying materials are made available
# under the terms of "Eclipse Public License v1.0"
# which accompanies this distribution, and is available
# at the URL "http://www.eclipse.org/legal/epl-v10.html".
#
# Initial Contributors:
# Nokia Corporation - initial contribution.
#
# Contributors:
#
# Description:
# Packaging framework main package
#

'''
A Packaging framework
=====================

Blocks is a framework for packaging files into easily distributed and managed
units. When a codebase changes, a correctly functioning packaging framework
provides a way to distribute only the units that have changed as a result. The
units are referred to as I{bundles}, rather than packages, in Blocks terminology
to avoid confusion with other uses of the term package.

Blocks is agnostic in terms of bundle formats, though the default Deb format
is the is best supported.

Blocks also provides separate end user tools for installing, upgrading and
removing packages created by the packaging framework.

Packaging framework usage - overview
====================================

L{Packager} provides the main interface for using the packager framework. The
following inputs are required for a Packager:
    - A L{PackagerStorage} object to provide paths to metadata files and build
    ID data to the Packager. Two implementations for different scenarios are
    provided:
        - L{DefaultPackagerStorage} maintains a I{component/build}
        directory hierarchy to store metadata and provides automatic build
        ID incrementation.
        - L{OneoffStorage} is used when a full build history is not required.
    - One or more L{BuildData} objects, each containing the necessary data to
    create a packageable component:
        - Component name, version
        - Lists of source and target files
        - Dependencies on other files
    The L{DataSources} module contains classes for creating BuildData from
    various types of input.
    - The output directory for written bundles.

Optional inputs include:
    - One or more names of previously packaged components, the metadata of which
    we wish to use for calculating dependencies.
    - L{DependencyProcessors} for creating dependencies between bundles.

Packaging can be tweaked by changing Packager variables, including specifying
an alternative L{PackageWriter} or packaging L{Rules} used to::
    - classify and allocate target and source files to bundles
    - name package files
    - create dependencies between packages within a component

Packaging framework usage - example
===================================
Import modules::
    import Blocks.Packaging

    # for reading Raptor output
    from Blocks.Packaging.DataSources.WhatLog import *
    from Ft.Xml.Xslt import Transform
    import xml.sax

    # dependency handling
    from Blocks.Packaging.DependencyProcessors import RaptorDependencyProcessor
    from Blocks.Packaging.DependencyProcessors.RomPatchProcessor import *

Read Raptor output from file I{input.xml}::
    parser = xml.sax.make_parser()
    reader = GenericLogReader()
    parser.setContentHandler(reader)
    parser.feed(Transform("input.xml", GenericLogReader.defaultStyleSheet))
    parser.close()

Use GenericLogReader to create BuildData from a list of files associated with
I{src/common/generic/my/group/bld.inf}::
    myBuildData = reader.createBuildData(
        infs=["/path/to/sourceroot/src/common/generic/my/group/bld.inf"],
        name="myComponent",
        version="1",
        )

    # add source data not provided by GenericLogReader
    myBuildData.addSourceFiles(["src/common/generic/my/group/bld.inf"])
    myBuildData.setSourceRoot("/path/to/sourceroot")

Create storage instance for storing package files and metadata in I{/tmp/myComponent}::
    # Use metadata from previous build, found in "/tmp/oldbuild/"
    storage = Blocks.Packaging.OneoffStorage(basedir="/tmp", build=2, previousMetaDir="/tmp/oldbuild/")

Create the packager::
    myPackager = Blocks.Packaging.Packager(storage, "/path/for/writing/packages")

Add dependency processors; indirect dependency rules are found at I{/foo/indirect.rules}::
    myPackager.addProcessor(RomPatchDependencyProcessor, None)
    myPackager.addProcessor(DotDeeDependencyProcessor, {})
    myPackager.addProcessor(Blocks.Packaging.BuildDataDependencyProcessor, "/foo/indirect.rules")

Read metadata of component I{otherComponent}, to which there may be dependencies::
    myPackager.addNonBuiltComponent("otherComponent")

Write build metadata and packages that have changed since the last build::
    myPackager.addComponent(myBuildData)
    myPackager.wait()
'''

class PackagingError(Exception):
    ''' Parent class for packaging framework exceptions '''

# Flat namespace with default classes

from Storage import DefaultPackagerStorage, OneoffStorage
from PackageModel import *
from BuildData import PlainBuildData, BdFile
from MultiprocessPackager import Packager
from Logging import Logging
from DependencyProcessors.DefaultProcessors import *
from PackageWriter import PackageWriter
import Rules